# Based on Wine source
# https://github.com/wine-mirror/wine/blob/master/dlls/kernelbase/locale.c

from enum import IntEnum, IntFlag

from dissect.util.xmemoryview import xmemoryview

from dissect.esedb.sorting_table import table


class MapFlags(IntFlag):
    NORM_IGNORECASE = 0x00000001  # ignore case
    NORM_IGNORENONSPACE = 0x00000002  # ignore nonspacing chars
    NORM_IGNORESYMBOLS = 0x00000004  # ignore symbols

    LINGUISTIC_IGNORECASE = 0x00000010  # linguistically appropriate 'ignore case'
    LINGUISTIC_IGNOREDIACRITIC = 0x00000020  # linguistically appropriate 'ignore nonspace'

    NORM_IGNOREKANATYPE = 0x00010000  # ignore kanatype
    NORM_IGNOREWIDTH = 0x00020000  # ignore width
    NORM_LINGUISTIC_CASING = 0x08000000  # use linguistic rules for casing

    LCMAP_LOWERCASE = 0x00000100  # lower case letters
    LCMAP_UPPERCASE = 0x00000200  # UPPER CASE LETTERS
    LCMAP_TITLECASE = 0x00000300  # Title Case Letters

    LCMAP_SORTKEY = 0x00000400  # WC sort key (normalize)
    LCMAP_BYTEREV = 0x00000800  # byte reversal

    LCMAP_HIRAGANA = 0x00100000  # map katakana to hiragana
    LCMAP_KATAKANA = 0x00200000  # map hiragana to katakana
    LCMAP_HALFWIDTH = 0x00400000  # map double byte to single byte
    LCMAP_FULLWIDTH = 0x00800000  # map single byte to double byte

    LCMAP_LINGUISTIC_CASING = 0x01000000  # use linguistic rules for casing

    LCMAP_SIMPLIFIED_CHINESE = 0x02000000  # map traditional chinese to simplified chinese
    LCMAP_TRADITIONAL_CHINESE = 0x04000000  # map simplified chinese to traditional chinese

    LCMAP_SORTHANDLE = 0x20000000
    LCMAP_HASH = 0x00040000

    SORT_STRINGSORT = 0x00001000  # use string sort method
    SORT_DIGITSASNUMBERS = 0x00000008  # use digits as numbers sort method


class SCRIPT(IntEnum):
    UNSORTABLE = 0
    NONSPACE_MARK = 1
    EXPANSION = 2
    EASTASIA_SPECIAL = 3
    JAMO_SPECIAL = 4
    EXTENSION_A = 5
    PUNCTUATION = 6
    SYMBOL_1 = 7
    SYMBOL_2 = 8
    SYMBOL_3 = 9
    SYMBOL_4 = 10
    SYMBOL_5 = 11
    SYMBOL_6 = 12
    DIGIT = 13
    LATIN = 14
    GREEK = 15
    CYRILLIC = 16
    KANA = 34
    HEBREW = 40
    ARABIC = 41
    PUA_FIRST = 169
    PUA_LAST = 175
    CJK_FIRST = 192
    CJK_LAST = 239


class CASE(IntFlag):
    FULLWIDTH = 0x01  # full width kana (vs. half width)
    FULLSIZE = 0x02  # full size kana (vs. small)
    SUBSCRIPT = 0x08  # sub/super script
    UPPER = 0x10  # upper case
    KATAKANA = 0x20  # katakana (vs. hiragana)
    COMPR_2 = 0x40  # compression exists for >= 2 chars
    COMPR_4 = 0x80  # compression exists for >= 4 chars
    COMPR_6 = 0xC0  # compression exists for >= 6 chars


def map_string(value: str, flags: MapFlags, locale: str) -> bytes:
    """Very basic Python implementation of LCMapStringEx, only supporting sorting keys.

    Currently only supports one hardcoded sorting table (the default) and the basic character types.

    Args:
        value: The string to map to a sorting key.
        flags: The flags passed to LCMapStringEx
        locale: The locale to use for mapping.

    Returns:
        A sorting key of the given input that should be compatible with LCMapStringEx.

    Raises:
        NotImplementedError: If an unsupported flag or character is encountered.
    """
    key_primary = []
    key_diacritic = []
    key_case = []

    if not flags & MapFlags.LCMAP_SORTKEY:
        raise NotImplementedError("Only LCMAP_SORTKEY is partially supported")

    case_mask = 0x3F
    if flags & MapFlags.NORM_IGNORECASE:
        case_mask &= ~(CASE.UPPER | CASE.SUBSCRIPT)
    if flags & MapFlags.NORM_IGNOREWIDTH:
        case_mask &= ~CASE.FULLWIDTH
    if flags & MapFlags.NORM_IGNOREKANATYPE:
        case_mask &= ~CASE.KATAKANA

    view = xmemoryview(value.encode("utf-16-le"), "<H")
    for cp in view:
        weight = table[cp]

        alphabetic_weight = weight >> 24
        script_member = (weight >> 16) & 0xFF
        diacritic_weight = (weight >> 8) & 0xFF
        case_weight = (weight & 0xFF) & case_mask

        if script_member == SCRIPT.UNSORTABLE:
            continue

        if script_member == SCRIPT.NONSPACE_MARK:
            if flags & MapFlags.LINGUISTIC_IGNOREDIACRITIC:
                diacritic_weight = 2

            if len(key_diacritic):
                key_diacritic[-1] += diacritic_weight
            else:
                key_diacritic.append(diacritic_weight)

        if script_member == SCRIPT.EXPANSION:
            raise NotImplementedError(SCRIPT.EXPANSION)

        if script_member == SCRIPT.EASTASIA_SPECIAL:
            raise NotImplementedError(SCRIPT.EASTASIA_SPECIAL)

        if script_member == SCRIPT.JAMO_SPECIAL:
            raise NotImplementedError(SCRIPT.JAMO_SPECIAL)

        if script_member == SCRIPT.EXTENSION_A:
            key_primary.append(0xFD)
            key_primary.append(0xFF)
            key_primary.append(alphabetic_weight)
            key_primary.append(diacritic_weight)
            key_diacritic.append(2)
            key_case.append(2)
            continue

        if script_member == SCRIPT.PUNCTUATION:
            if flags & MapFlags.NORM_IGNORESYMBOLS:
                continue

            if not flags & MapFlags.SORT_STRINGSORT:
                raise NotImplementedError(SCRIPT.PUNCTUATION)

            key_primary.append(script_member)
            key_primary.append(alphabetic_weight)
            key_diacritic.append(diacritic_weight)
            key_case.append(case_weight)
            continue

        if script_member in (
            SCRIPT.SYMBOL_1,
            SCRIPT.SYMBOL_2,
            SCRIPT.SYMBOL_3,
            SCRIPT.SYMBOL_4,
            SCRIPT.SYMBOL_5,
            SCRIPT.SYMBOL_6,
        ):
            if flags & MapFlags.NORM_IGNORESYMBOLS:
                continue
            key_primary.append(script_member)
            key_primary.append(alphabetic_weight)
            key_diacritic.append(diacritic_weight)
            key_case.append(case_weight)
            continue

        if script_member == SCRIPT.DIGIT and 0:
            raise NotImplementedError(SCRIPT.DIGIT)

        # else
        key_primary.append(script_member)
        key_primary.append(alphabetic_weight)
        key_diacritic.append(diacritic_weight)
        key_case.append(case_weight)

    key_diacritic = _filter_weights(key_diacritic)
    if flags & (MapFlags.NORM_IGNORECASE | MapFlags.NORM_IGNOREWIDTH):
        key_case = []
    else:
        key_case = _filter_weights(key_case)

    return bytes(
        [
            *key_primary,
            0x01,
            *(key_diacritic if not flags & MapFlags.NORM_IGNORENONSPACE else []),
            0x01,
            *key_case,
            0x01,
            # extra would go here
            0x01,
            # special would go here
            0x00,
        ]
    )


def _filter_weights(weights):
    i = len(weights)
    while i > 0:
        if weights[i - 1] > 2:
            break
        i -= 1
    return weights[:i]



import gzip
import os

import pytest


def absolute_path(filename):
    return os.path.join(os.path.dirname(__file__), filename)


def open_file(name, mode="rb"):
    with open(absolute_path(name), mode) as f:
        yield f


def open_file_gz(name, mode="rb"):
    with gzip.GzipFile(absolute_path(name), mode) as f:
        yield f


@pytest.fixture
def basic_db():
    yield from open_file_gz("data/basic.edb.gz")


@pytest.fixture
def binary_db():
    yield from open_file_gz("data/binary.edb.gz")


@pytest.fixture
def text_db():
    yield from open_file_gz("data/text.edb.gz")


@pytest.fixture
def multi_db():
    yield from open_file_gz("data/multi.edb.gz")


@pytest.fixture
def default_db():
    yield from open_file_gz("data/default.edb.gz")


@pytest.fixture
def index_db():
    yield from open_file_gz("data/index.edb.gz")


@pytest.fixture
def large_db():
    yield from open_file_gz("data/large.edb.gz")


@pytest.fixture
def sru_db():
    yield from open_file_gz("data/SRUDB.dat.gz")


@pytest.fixture
def ual_db():
    yield from open_file_gz("data/Current.mdb.gz")



from __future__ import annotations

from typing import TYPE_CHECKING, Union

from dissect.esedb.exceptions import KeyNotFoundError, NoNeighbourPageError
from dissect.esedb.page import Node, Page

if TYPE_CHECKING:
    from dissect.esedb.esedb import EseDB


class Cursor:
    """A simple cursor implementation for searching the ESE B+Trees

    Args:
        esedb: An instance of :class:`~dissect.esedb.esedb.EseDB`.
        page: The page to open a cursor on.
    """

    def __init__(self, esedb: EseDB, page: Union[int, Page]):
        self.esedb = esedb

        if isinstance(page, int):
            page_num = page
            page = esedb.page(page_num)
        else:
            page_num = page.num

        self._page = page
        self._page_num = page_num
        self._node_num = 0

    def node(self) -> Node:
        """Return the node the cursor is currently on."""
        return self._page.node(self._node_num)

    def next(self) -> Node:
        """Move the cursor to the next node and return it.

        Can move the cursor to the next page as a side effect.
        """
        if self._node_num + 1 > self._page.node_count - 1:
            self.next_page()
        else:
            self._node_num += 1

        return self.node()

    def next_page(self) -> None:
        """Move the cursor to the next page in the tree.

        Raises:
            NoNeighbourPageError: If the current page has no next page.
        """
        if self._page.next_page:
            self._page = self.esedb.page(self._page.next_page)
            self._node_num = 0
        else:
            raise NoNeighbourPageError(f"{self._page} has no next page")

    def prev(self) -> Node:
        """Move the cursor to the previous node and return it.

        Can move the cursor to the previous page as a side effect.
        """
        if self._node_num - 1 < 0:
            self.prev_page()
        else:
            self._node_num -= 1

        return self.node()

    def prev_page(self) -> None:
        """Move the cursor to the previous page in the tree.

        Raises:
            NoNeighbourPageError: If the current page has no previous page.
        """
        if self._page.previous_page:
            self._page = self.esedb.page(self._page.previous_page)
            self._node_num = self._page.node_count - 1
        else:
            raise NoNeighbourPageError(f"{self._page} has no previous page")

    def search(self, key: bytes, exact: bool = True) -> Node:
        """Search the tree for the given key.

        Moves the cursor to the matching node, or on the last node that is less than the requested key.

        Args:
            key: The key to search for.
            exact: Whether to only return successfully on an exact match.

        Raises:
            KeyNotFoundError: If an ``exact`` match was requested but not found.
        """
        page = self._page
        while True:
            node = find_node(page, key)

            if page.is_branch:
                page = self.esedb.page(node.child)
            else:
                self._page = page
                self._page_num = page.num
                self._node_num = node.num
                break

        if exact and key != node.key:
            raise KeyNotFoundError(f"Can't find key: {key}")

        return self.node()


def find_node(page: Page, key: bytes) -> Node:
    """Search the tree, starting from the given ``page`` and search for ``key``.

    Args:
        page: The page to start searching from. Should be a branch page.
        key: The key to search.
    """
    first_node_idx = 0
    last_node_idx = page.node_count - 1

    node = None
    while first_node_idx < last_node_idx:
        node_idx = (first_node_idx + last_node_idx) // 2
        node = page.node(node_idx)

        # It turns out that the way BTree keys are compared matches 1:1 with how Python compares bytes
        # First compare data, then length
        if key < node.key:
            last_node_idx = node_idx
        elif key == node.key:
            if page.is_branch:
                # If there's an exact match on a key on a branch page, the actual leaf nodes are in the next branch
                # Page keys for branch pages appear to be non-inclusive upper bounds
                node_idx = min(node_idx + 1, page.node_count - 1)
                node = page.node(node_idx)

            return node
        else:
            first_node_idx = node_idx + 1

    # We're at the last node
    return page.node(first_node_idx)



from dissect.esedb.esedb import EseDB
from dissect.esedb.exceptions import (
    Error,
    InvalidDatabase,
    KeyNotFoundError,
    NoNeighbourPageError,
)

__all__ = [
    "EseDB",
    "Error",
    "CompressedTaggedDataError",
    "InvalidDatabase",
    "KeyNotFoundError",
    "NoNeighbourPageError",
]



from typing import BinaryIO

from dissect.esedb.esedb import EseDB


def test_as_dict(basic_db: BinaryIO):
    db = EseDB(basic_db)
    table = db.table("basic")

    records = list(table.records())
    assert len(records) == 2

    assert [r.as_dict() for r in records] == [
        {
            "Id": 1,
            "Bit": False,
            "UnsignedByte": 213,
            "Short": -1337,
            "Long": -13371337,
            "Currency": 1337133713371337,
            "IEEESingle": 1.0,
            "IEEEDouble": 13371337.13371337,
            "DateTime": 4675210852477960192,
            "UnsignedLong": 13371337,
            "LongLong": -13371337,
            "GUID": "3f360af1-6766-46dc-9af2-0dacf295c2a1",
            "UnsignedShort": 1337,
        },
        {
            "Id": 2,
            "Bit": True,
            "UnsignedByte": 255,
            "Short": 1339,
            "Long": 13391339,
            "Currency": -1339133913391339,
            "IEEESingle": -2.0,
            "IEEEDouble": -13391339.13391339,
            "DateTime": -4537072128574357504,
        },
    ]



from dissect.esedb.esedb import EseDB


def test_index(index_db):
    db = EseDB(index_db)
    table = db.table("index")

    assert len(table.indexes) == 19

    assert table.indexes[0].name == "IxId"
    assert table.indexes[0].column_ids == [1]
    assert table.indexes[0].search(Id=1)

    assert table.indexes[1].name == "IxBit"
    assert table.indexes[1].column_ids == [2]
    assert table.indexes[1].search(Bit=False)

    assert table.indexes[2].name == "IxUnsignedByte"
    assert table.indexes[2].column_ids == [3]
    assert table.indexes[2].search(UnsignedByte=213)

    assert table.indexes[3].name == "IxShort"
    assert table.indexes[3].column_ids == [4]
    assert table.indexes[3].search(Short=-1337)

    assert table.indexes[4].name == "IxLong"
    assert table.indexes[4].column_ids == [5]
    assert table.indexes[4].search(Long=-13371337)

    assert table.indexes[5].name == "IxCurrenc"
    assert table.indexes[5].column_ids == [6]
    assert table.indexes[5].search(Currency=1337133713371337)

    assert table.indexes[6].name == "IxIEEESingle"
    assert table.indexes[6].column_ids == [7]
    assert table.indexes[6].search(IEEESingle=1.0)

    assert table.indexes[7].name == "IxIEEEDouble"
    assert table.indexes[7].column_ids == [8]
    assert table.indexes[7].search(IEEEDouble=13371337.13371337)

    assert table.indexes[8].name == "IxDateTime"
    assert table.indexes[8].column_ids == [9]
    assert table.indexes[8].search(DateTime=4675210852477960192)

    assert table.indexes[9].name == "IxUnsignedLong"
    assert table.indexes[9].column_ids == [10]
    assert table.indexes[9].search(UnsignedLong=13371337)

    assert table.indexes[10].name == "IxLongLong"
    assert table.indexes[10].column_ids == [11]
    assert table.indexes[10].search(LongLong=-13371337)

    assert table.indexes[11].name == "IxGUID"
    assert table.indexes[11].column_ids == [12]
    assert table.indexes[11].search(GUID="3f360af1-6766-46dc-9af2-0dacf295c2a1")

    assert table.indexes[12].name == "IxUnsignedShort"
    assert table.indexes[12].column_ids == [13]
    assert table.indexes[12].search(UnsignedShort=1337)

    assert table.indexes[13].name == "IxBinary"
    assert table.indexes[13].column_ids == [128]
    assert table.indexes[13].search(Binary=b"test binary data")

    assert table.indexes[14].name == "IxLongBinary"
    assert table.indexes[14].column_ids == [256]
    assert table.indexes[14].search(LongBinary=b"test long binary data " + (b"a" * 1024))

    assert table.indexes[15].name == "IxASCII"
    assert table.indexes[15].column_ids == [129]
    assert table.indexes[15].search(ASCII="Simple ASCII text")

    assert table.indexes[16].name == "IxUnicode"
    assert table.indexes[16].column_ids == [130]
    assert table.indexes[16].search(Unicode="Simple Unicode text 🦊")

    assert table.indexes[17].name == "IxLongASCII"
    assert table.indexes[17].column_ids == [257]
    assert table.indexes[17].search(LongASCII="Long ASCII text " + ("a" * 1024))

    assert table.indexes[18].name == "IxLongUnicode"
    assert table.indexes[18].column_ids == [258]
    assert table.indexes[18].search(LongUnicode="Long Unicode text 🦊 " + ("a" * 1024))



from __future__ import annotations

import argparse
from typing import BinaryIO, Iterator, Optional

from dissect.util.sid import read_sid
from dissect.util.ts import oatimestamp, wintimestamp

from dissect.esedb.c_esedb import RecordValue
from dissect.esedb.esedb import EseDB
from dissect.esedb.record import Record, serialise_record_column_values
from dissect.esedb.table import Table

NATIVE_TYPE_MAP = {
    "{DD6636C4-8929-4683-974E-22C046A43763}": {"ConnectStartTime": wintimestamp},
    "{5C8CF1C7-7257-4F13-B223-970EF5939312}": {"EndTime": wintimestamp},
}

SKIP_TABLES = [
    "MSysObjects",
    "MSysObjectsShadow",
    "MSysObjids",
    "MSysLocales",
    "SruDbIdMapTable",
    "SruDbCheckpointTable",
]


NAME_TO_GUID_MAP = {
    "network_data": "{973F5D5C-1D90-4944-BE8E-24B94231A174}",
    "network_connectivity": "{DD6636C4-8929-4683-974E-22C046A43763}",
    "energy_estimator": "{DA73FB89-2BEA-4DDC-86B8-6E048C6DA477}",
    "energy_usage": "{FEE4E14F-02A9-4550-B5CE-5FA2DA202E37}",
    "energy_usage_lt": "{FEE4E14F-02A9-4550-B5CE-5FA2DA202E37}LT",
    "application": "{D10CA2FE-6FCF-4F6D-848E-B2E99266FA89}",
    "push_notifications": "{D10CA2FE-6FCF-4F6D-848E-B2E99266FA86}",
    "application_timeline": "{5C8CF1C7-7257-4F13-B223-970EF5939312}",
    "vfu": "{7ACBBAA3-D029-4BE4-9A7A-0885927F1D8F}",
    # Refs:
    # - http://dfir.pro/index.php?link_id=92259 and
    # - HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion\SRUM\Extensions
    "sdp_volume_provider": "{17F4D97B-F26A-5E79-3A82-90040A47D13D}",
    "sdp_physical_disk_provider": "{841A7317-3805-518B-C2EA-AD224CB4AF84}",
    "sdp_cpu_provider": "{DC3D3B50-BB90-5066-FA4E-A5F90DD8B677}",
    "sdp_network_provider": "{EEE2F477-0659-5C47-EF03-6D6BEFD441B3}",
    # not seen in the wild
    "sdp_perf_count_provider": "{38AD6548-9313-58F8-45C7-D293BAFDC879}",
    "sdp_event_log_provider": "{CDF8EBF6-7C0F-5AC2-158F-DBFBEE981152}",
}


class SRU:
    def __init__(self, fh: BinaryIO):
        self.esedb = EseDB(fh)

        id_map_table = self.esedb.table("SruDbIdMapTable")
        self.id_map = {r.get("IdIndex"): r for r in id_map_table.records()}

    def get_table(self, table_name: str = None, table_guid: str = None) -> Optional[Table]:
        if all((table_name, table_guid)) or not any((table_name, table_guid)):
            raise ValueError("Either table_name or table_guid must be provided")

        if table_name and table_name not in NAME_TO_GUID_MAP:
            raise ValueError(f"Unknown table name: {table_name}")

        table_guid = table_guid or NAME_TO_GUID_MAP[table_name]

        try:
            return self.esedb.table(table_guid)
        except KeyError:
            return None

    def entries(self) -> Iterator[Entry]:
        for t in self.esedb.tables():
            if t.name in SKIP_TABLES:
                continue
            yield from self.get_table_entries(table=t)

    __iter__ = entries

    def get_table_entries(self, table: Table = None, table_name: str = None, table_guid: str = None) -> Iterator[Entry]:
        table = table or self.get_table(table_name=table_name, table_guid=table_guid)
        if not table:
            return
        for record in table.records():
            yield Entry(self, table, record)

    def resolve_id(self, value: int) -> Optional[str]:
        try:
            record = self.id_map[value]
        except KeyError:
            raise IndexError(value)

        if not record.get("IdBlob"):
            return None

        if record.get("IdType") in (0, 1, 2):
            return record.get("IdBlob").decode("utf-16-le").rstrip("\x00")
        else:
            return read_sid(record.get("IdBlob"))


class Entry:
    def __init__(self, sru: SRU, table: Table, record: Record):
        self.sru = sru
        self.table = table
        self.record = record
        self.provider = table.name

    def _get(self, attr: str) -> RecordValue:
        value = self.record.get(attr)
        if value is None:
            return value

        if attr in ("AppId", "UserId"):
            value = self.sru.resolve_id(value)

        if attr == "TimeStamp":
            value = oatimestamp(value)

        type_map = NATIVE_TYPE_MAP.get(self.table.name, None)
        if type_map and attr in type_map:
            value = type_map[attr](value)

        # Application timeline ({5C8CF1C7-7257-4F13-B223-970EF5939312}) seems to have a weird null value
        # (******** in decimal)
        if self.table.name == "{5C8CF1C7-7257-4F13-B223-970EF5939312}" and value in (0x2A2A2A2A2A2A2A2A, 0x2A2A2A2A):
            value = None

        return value

    def __getitem__(self, attr: str) -> RecordValue:
        return self._get(attr)

    def __getattr__(self, attr: str) -> RecordValue:
        try:
            return self._get(attr)
        except Exception:
            return object.__getattribute__(self, attr)

    def __repr__(self) -> str:
        column_values = serialise_record_column_values(self.record)
        return f"<Entry provider={self.table.name} {column_values}>"


def main():
    parser = argparse.ArgumentParser(description="dissect.esedb SRU parser")
    parser.add_argument("input", help="SRU database to read")
    parser.add_argument("-p", "--provider", help="filter records from this provider")
    args = parser.parse_args()

    with open(args.input, "rb") as fh:
        parser = SRU(fh)

        if args.provider in NAME_TO_GUID_MAP:
            for e in parser.get_table_entries(table_name=args.provider):
                print(e)
        elif args.provider:
            for e in parser.get_table_entries(table_guid=args.provider):
                print(e)
        else:
            for e in parser.entries():
                print(e)


if __name__ == "__main__":
    main()



from __future__ import annotations

import functools
import struct
from binascii import hexlify
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Iterator, Optional

from dissect.util.xmemoryview import xmemoryview

from dissect.esedb import compression
from dissect.esedb.c_esedb import PAGE_FLAG, TAGFLD_HEADER, RecordValue, c_esedb

if TYPE_CHECKING:
    from dissect.esedb.page import Node
    from dissect.esedb.table import Column, Table


def noop(value: Any):
    return value


class Record:
    """Wrapper class for records in a table.

    The actual parsing of the data is done in :class:`RecordData`, but this class allows you to easily
    retrieve all the values by either using the `.get()` method, accessing them as attributes or dictionary keys
    on this class.

    Args:
        table: The table this record is from.
        node: The node of this record.
    """

    def __init__(self, table: Table, node: Node):
        self._table = table
        self._esedb = table.esedb
        self._node = node
        self._data = RecordData(table, node)

    def get(self, attr: str, raw: bool = False) -> RecordValue:
        """Retrieve a value from the record with the given name.

        Optionally receive the raw data as it's stored in the record.

        Args:
            attr: The column name to retrieve the value of.
            raw: Whether to return the raw data stored in the record instead of the parsed value.
        """
        column = self._table.column(attr)
        return self._data.get(column, raw)

    def as_dict(self, raw: bool = False) -> dict[str, RecordValue]:
        return self._data.as_dict(raw)

    def __getitem__(self, attr: str) -> RecordValue:
        return self.get(attr)

    def __getattr__(self, attr: str) -> RecordValue:
        try:
            return self.get(attr)
        except KeyError:
            return object.__getattribute__(self, attr)

    def __str__(self) -> str:
        column_values = serialise_record_column_values(self, max_columns=None)
        return f"<Record {column_values}>"

    def __repr__(self) -> str:
        column_values = serialise_record_column_values(self)
        return f"<Record {column_values}>"


class RecordData:
    """Record class for parsing and interacting with the on-disk record format.

    Templated columns are currently not implemented.

    Args:
        table: The table this record is from.
        data: The node data of this record.

    Raises:
        NotImplementedError: If old format tagged fields are encountered.
    """

    def __init__(self, table: Table, node: Node):
        self.table = table
        self.esedb = table.esedb
        self.node = node
        self.data = node.data

        self.header = None
        self._values = {}

        self._last_fixed_id = None
        self._last_variable_id = None

        self._fixed_null_bitmap = None

        self._variable_offset_start = None
        self._variable_data_start = None
        self._variable_offsets = []

        self._tagged_data_start = None
        self._tagged_data_count = 0
        self._tagged_data_view = None
        self._tagged_fields = {}

        if len(self.data) >= 4:
            self.header = c_esedb.RECHDR(self.data)
            self._last_fixed_id = self.header.fidFixedLastInRec
            self._last_variable_id = self.header.fidVarLastInRec
            self._variable_offset_start = self.header.ibEndOfFixedData

            # There's a bitmap between the end of the fixed data and the start of variable data that indicates
            # if a fixed column is null.
            bitmap_start = self._variable_offset_start - ((self._last_fixed_id + 7) // 8)
            self._fixed_null_bitmap = self.data[bitmap_start : self._variable_offset_start]

            # Calculate where the variable offsets array and data start
            # This info is needed for parsing both variable and tagged data
            # Variable identifiers start from 128
            num_variable = self._last_variable_id - 127
            # Variable offset end == variable data start
            self._variable_data_start = self._variable_offset_start + (num_variable * 2)

            if num_variable > 0 and len(self.data) >= 4 + (num_variable * 2):
                # Parse the variable offsets already, if we have them
                # There can only be 128 at most, so this shouldn't be an expensive operation
                self._variable_offsets = struct.unpack(
                    "<%dH" % num_variable, self.data[self._variable_offset_start : self._variable_data_start]
                )

            self._tagged_data_start = self._variable_data_start
            if self._variable_offsets:
                self._tagged_data_start += self._variable_offsets[-1] & 0x7FFF

            if len(self.data) >= self._tagged_data_start + 4:
                if node.tag.page.flags & PAGE_FLAG.NewRecordFormat == 0:
                    raise NotImplementedError("Record has tagged fields in an old format, which is not implemented yet")

                tag_value = int.from_bytes(self.data[self._tagged_data_start : self._tagged_data_start + 4], "little")
                first_tagged_field = TagField(self, tag_value)

                tagged_field_data_start = self._tagged_data_start
                tagged_field_data_end = tagged_field_data_start + first_tagged_field.offset
                tagged_field_data = self.data[self._tagged_data_start : tagged_field_data_end]

                self._tagged_data_count = first_tagged_field.offset // 4  # sizeof(TAGFLD)
                self._tagged_data_view = xmemoryview(tagged_field_data, "<I")
                self._tagged_fields[first_tagged_field.identifier] = first_tagged_field

    def get(self, column: Column, raw: bool = False) -> RecordValue:
        """Retrieve the value for the specified column.

        Optionally receive the raw data as it's stored in the record.

        If the database has been opened in impacket compatibility mode, skip most of the parsing and return the values
        that impacket expects.

        Args:
            column: The column to retrieve the value of.
            raw: Whether to return the raw data stored in the record instead of the parsed value.
        """
        value = None
        tag_field = None

        if self.header is None:
            return value

        if column.is_fixed:
            value = self._get_fixed(column)

        elif column.is_variable:
            value = self._get_variable(column)

        elif column.is_tagged:
            tag_field, value = self._get_tagged(column)

        if raw:
            if isinstance(value, memoryview):
                value = value.tobytes()
            return value

        if value is not None:
            return self._parse_value(column, value, tag_field)

    def as_dict(self, raw: bool = False) -> dict[str, RecordValue]:
        """Serialize the record as a dictionary."""
        obj = {}

        def _iter_column_id() -> Iterator[Column]:
            # Fixed
            yield from range(1, self._last_fixed_id + 1)

            # Variable
            yield from range(128, self._last_variable_id + 1)

            # Tagged
            for idx in range(self._tagged_data_count):
                yield self._get_tag_field(idx).identifier

        for column_id in _iter_column_id():
            column = self.table._column_id_map[column_id]

            try:
                obj[column.name] = self.get(column, raw)
            except Exception as e:
                obj[column.name] = f"!ERROR! {e}"

        return obj

    def _parse_value(self, column: Column, value: bytes, tag_field: TagField = None) -> RecordValue:
        """Parse the raw value into the appropriate type.

        For tagged columns, also interpret things like multi-values, separated and compressed data.
        """
        ctype = column.ctype
        parse_func = ctype.parse
        if column.is_text:
            parse_func = functools.partial(ctype.parse, encoding=column.encoding)

        if self.esedb.impacket_compat:
            if tag_field and tag_field.flags & TAGFLD_HEADER.Compressed:
                value = None
            elif tag_field and tag_field.flags & TAGFLD_HEADER.MultiValues:
                value = hexlify(value)
            elif parse_func != bytes:
                value = parse_func(value)
            else:
                value = hexlify(value)
        else:
            if tag_field:
                if tag_field.flags & TAGFLD_HEADER.MultiValues:
                    value = self._parse_multivalue(value, tag_field)
                else:
                    if tag_field.flags & TAGFLD_HEADER.Separated:
                        value = self.table.get_long_value(bytes(value))
                    elif tag_field.flags & TAGFLD_HEADER.Compressed:
                        # Long values are already decompressed during retrieval
                        value = compression.decompress(value)

            parse_func = parse_func or noop
            if tag_field and tag_field.flags & TAGFLD_HEADER.MultiValues:
                value = list(map(parse_func, value))
            else:
                value = parse_func(value)

        return value

    def _parse_multivalue(self, value: bytes, tag_field: TagField):
        fSeparatedInstance = 0x8000

        if tag_field.flags & TAGFLD_HEADER.TwoValues:
            # Optimized storage for when a multi-value only has two values
            # First byte is the size of the first value, calculate the size of the second value from that
            first_size = value[0]
            second_size = len(value) - (1 + first_size)
            value = [value[1 : 1 + first_size], value[1 + first_size : 1 + first_size + second_size]]
        elif tag_field.flags & TAGFLD_HEADER.MultiValues:
            # Regular multi-value storage, starts with an array of USHORT offsets to the actual values
            # Just calculate the amount of values from the first entry
            # Individual offsets can have a fSeparatedInstance (0x8000) flag set
            first_value_offset = struct.unpack("<H", value[0:2])[0] & 0x7FFF
            num_values = first_value_offset // 2  # sizeof(USHORT)
            value_offsets = struct.unpack(f"<{num_values}H", value[:first_value_offset]) + (len(value),)

            values = []
            for i in range(num_values):
                offset = value_offsets[i]
                data = value[offset & 0x7FFF : value_offsets[i + 1] & 0x7FFF]
                if offset & fSeparatedInstance:
                    data = self.table.get_long_value(bytes(data))
                values.append(data)
            value = values

        if tag_field.flags & TAGFLD_HEADER.Compressed:
            # Only the first entry appears to be compressed
            value[0] = compression.decompress(value[0])

        return value

    def _get_fixed(self, column: Column) -> Optional[bytes]:
        """Parse a specific fixed column."""
        if column.identifier <= self._last_fixed_id:
            # Check if it's not null
            bit_idx_identifier = column.identifier - 1
            bitmap_offset, bitmap_shift = divmod(bit_idx_identifier, 8)
            if self._fixed_null_bitmap[bitmap_offset] & (1 << bitmap_shift):
                return None

            # Fixed data starts right after the header, which is 4 bytes
            offset = 4 + column.offset
            value = self.data[offset : offset + column.size]
        else:
            # If the column has a default, use that
            # If not, this defaults to None
            value = column.default

        return value

    def _get_variable(self, column: Column) -> Optional[bytes]:
        """Parse a specific variable column."""
        if column.identifier <= self._last_variable_id:
            identifier_idx = column.identifier - 128
            if identifier_idx == 0:
                value_start = 0
            else:
                # Start of this value is the end of the previous value
                # Even empty values have the offset encoded in them
                value_start = self._variable_offsets[identifier_idx - 1] & 0x7FFF

            # The value at the own index is the end offset of this value
            value_end = self._variable_offsets[identifier_idx]

            # If the MSB has been set, it means the entry is empty
            if value_end & 0x8000 == 0:
                # Offset everything with the variable data value starting offset
                value_offset = self._variable_data_start
                value = self.data[value_offset + value_start : value_offset + value_end]
            else:
                value = None
        else:
            # If the column has a default, use that
            # If not, this defaults to None
            value = column.default

        return value

    def _get_tagged(self, column: Column) -> Optional[bytes]:
        """Parse a specific tagged column."""
        tag_field = None

        idx = self._find_tag_field_idx(column.identifier)
        if idx is not None:
            tag_field = self._get_tag_field(idx)

            data_start = tag_field.offset
            if tag_field.has_extended_info:
                data_start += 1

            if idx + 1 < self._tagged_data_count:
                data_end = self._get_tag_field(idx + 1).offset
            else:
                data_end = len(self.data)

            if not tag_field.is_null:
                offset = self._tagged_data_start
                value = self.data[offset + data_start : offset + data_end]
        else:
            # If the column has a default, use that
            # If not, this defaults to None
            value = column.default

        return tag_field, value

    @lru_cache(4096)
    def _get_tag_field(self, idx: int) -> TagField:
        """Retrieve the :class:`TagField` at the given index in the ``TAGFLD`` array."""
        return TagField(self, self._tagged_data_view[idx])

    @lru_cache(4096)
    def _find_tag_field_idx(self, identifier: int, is_derived: bool = False) -> Optional[TagField]:
        """Find a tag field by identifier and optional derived flag.

        Performs a binary search in the tagged field array for the given identifier. The comparison algorithm used is
        derived from the ``TAGFLD::CmpTagfld2`` function in the Microsoft codebase.

        Args:
            identifier: The column identifier to look for.
            is_derived: A flag indicating if the column is derived.

        Returns:
            A :class:`TagField` object for the found tag field, or ``None`` if it's not found.
        """
        if self._tagged_data_count == 0:
            return None

        lookup = identifier
        if is_derived:
            lookup |= TagField.fDerived << 16  # fDerived

        # TAGFLD::CmpTagfld2
        flip_derived = TagField.fDerived << 16
        mask = flip_derived | ((1 << 16) - 1)
        value2 = flip_derived ^ (lookup & mask)

        min_idx = 0
        max_idx = self._tagged_data_count - 1

        while min_idx != max_idx:
            test_idx = min_idx + (max_idx - min_idx) // 2

            tag_value = self._tagged_data_view[test_idx]
            value1 = flip_derived ^ (tag_value & mask)
            if value1 < value2:
                min_idx = test_idx + 1
            elif value1 == value2:
                min_idx = test_idx
                break
            else:
                max_idx = test_idx

        tag_value = self._tagged_data_view[min_idx]
        if tag_value & 0xFFFF == identifier:
            return min_idx

        return None


class TagField:
    """Represents a ``TAGFLD``, which contains information about a tagged field in a record."""

    __slots__ = ("record", "identifier", "_offset", "offset", "has_extended_info", "flags")

    fNullSmallPage = 0x2000
    fDerived = 0x8000

    def __init__(self, record: RecordData, value: int):
        self.record = record
        self.identifier = value & 0xFFFF
        self._offset = (value >> 16) & 0xFFFF

        if record.esedb.has_small_pages:
            self.offset = self._offset & 0x1FFF  # & maskIb
            self.has_extended_info = bool(self._offset & 0x4000)  # & fExtendedInfo
        else:
            self.offset = self._offset & 0x7FFF  # & maskIb
            self.has_extended_info = True

        if self.has_extended_info and len(self.record.data) >= self.record._tagged_data_start + self.offset:
            self.flags = TAGFLD_HEADER(self.record.data[self.record._tagged_data_start + self.offset])
        else:
            self.flags = TAGFLD_HEADER.Invalid  # Made up flag member to keep the types consistent

    def __repr__(self) -> str:
        return f"<TagField identifier={self.identifier} offset={self.offset:#x} flags={str(self.flags).split('.')[1]}>"

    @property
    def is_null(self) -> bool:
        """Return whether this tagged field is null."""
        if self.record.esedb.has_small_pages:
            return bool(self._offset & TagField.fNullSmallPage)
        else:
            return bool(self.flags & TAGFLD_HEADER.Null)

    @property
    def is_derived(self) -> bool:
        """Return whether this tagged field is derived."""
        return bool(self._offset & TagField.fDerived)


def serialise_record_column_values(record: Record, column_names: list[str] = None, max_columns: int = 10) -> str:
    column_names = column_names or record._table.column_names
    columns_with_values = []
    for name in column_names:
        try:
            value = record.get(name)
        except Exception:
            value = "!ERROR!"

        if value is not None:
            columns_with_values.append((name, value))

        if max_columns and len(columns_with_values) >= max_columns:
            break

    reprs = " ".join([f"{name}={value!r}" for (name, value) in columns_with_values])
    has_more = " ..." if max_columns and len(column_names) > max_columns else ""
    return f"{reprs}{has_more}"



import struct
from typing import Optional

from dissect.util.compression import lzxpress, sevenbit

from dissect.esedb.c_esedb import COMPRESSION_SCHEME


def decompress(buf: bytes) -> bytes:
    """Decompress the given bytes according to the encoded compression scheme.

    Args:
        buf: The compressed bytes to decompress.

    Raises:
        NotImplementedError: If the buffer is compressed with an unsupported compression algorithm (XPRESS9/XPRESS10).
    """
    identifier = buf[0] >> 3
    if identifier == COMPRESSION_SCHEME.COMPRESS_7BITASCII:
        return sevenbit.decompress(buf[1:])
    elif identifier == COMPRESSION_SCHEME.COMPRESS_7BITUNICODE:
        return sevenbit.decompress(buf[1:], wide=True)
    elif identifier == COMPRESSION_SCHEME.COMPRESS_XPRESS:
        return lzxpress.decompress(buf[3:])
    elif identifier in (COMPRESSION_SCHEME.COMPRESS_XPRESS9, COMPRESSION_SCHEME.COMPRESS_XPRESS10):
        raise NotImplementedError(f"Compression not yet implemented: {COMPRESSION_SCHEME(identifier)}")
    else:
        # Not compressed
        return buf


def decompress_size(buf: bytes) -> Optional[int]:
    """Return the decompressed size of the given bytes according to the encoded compression scheme.

    Args:
        buf: The compressed bytes to return the decompressed size of.

    Raises:
        NotImplementedError: If the buffer is compressed with an unsupported compression algorithm (XPRESS9/XPRESS10).
    """
    identifier = buf[0] >> 3
    if identifier == COMPRESSION_SCHEME.COMPRESS_7BITASCII:
        return ((buf[0] & 7) + (8 * len(buf))) // 7
    elif identifier == COMPRESSION_SCHEME.COMPRESS_7BITUNICODE:
        return 2 * (((buf[0] & 7) + (8 * len(buf))) // 7)
    elif identifier == COMPRESSION_SCHEME.COMPRESS_XPRESS:
        return struct.unpack("<H", buf[1:2])[0]
    elif identifier in (COMPRESSION_SCHEME.COMPRESS_XPRESS9, COMPRESSION_SCHEME.COMPRESS_XPRESS10):
        raise NotImplementedError(f"Compression not yet implemented: {COMPRESSION_SCHEME(identifier)}")
    return None



import datetime

from dissect.util.ts import oatimestamp

from dissect.esedb.c_esedb import JET_coltyp
from dissect.esedb.esedb import EseDB


def test_basic_types(basic_db):
    db = EseDB(basic_db)
    table = db.table("basic")

    assert [(col.name, col.type) for col in table.columns] == [
        ("Id", JET_coltyp.Long),
        ("Bit", JET_coltyp.Bit),
        ("UnsignedByte", JET_coltyp.UnsignedByte),
        ("Short", JET_coltyp.Short),
        ("Long", JET_coltyp.Long),
        ("Currency", JET_coltyp.Currency),
        ("IEEESingle", JET_coltyp.IEEESingle),
        ("IEEEDouble", JET_coltyp.IEEEDouble),
        ("DateTime", JET_coltyp.DateTime),
        ("UnsignedLong", JET_coltyp.UnsignedLong),
        ("LongLong", JET_coltyp.LongLong),
        ("GUID", JET_coltyp.GUID),
        ("UnsignedShort", JET_coltyp.UnsignedShort),
    ]

    records = list(table.records())
    assert len(records) == 2

    assert records[0].Id == 1
    assert records[0].Bit is False
    assert records[0].UnsignedByte == 213
    assert records[0].Short == -1337
    assert records[0].Long == -13371337
    assert records[0].Currency == 1337133713371337
    assert records[0].IEEESingle == 1.0
    assert records[0].IEEEDouble == 13371337.13371337
    assert oatimestamp(records[0].DateTime) == datetime.datetime(1999, 3, 1, 0, 0, tzinfo=datetime.timezone.utc)
    assert records[0].UnsignedLong == 13371337
    assert records[0].LongLong == -13371337
    assert records[0].GUID == "3f360af1-6766-46dc-9af2-0dacf295c2a1"
    assert records[0].UnsignedShort == 1337

    assert records[1].Id == 2
    assert records[1].Bit is True
    assert records[1].UnsignedByte == 255
    assert records[1].Short == 1339
    assert records[1].Long == 13391339
    assert records[1].Currency == -1339133913391339
    assert records[1].IEEESingle == -2.0
    assert records[1].IEEEDouble == -13391339.13391339
    assert oatimestamp(records[1].DateTime) == datetime.datetime(1337, 6, 9, 0, 0, tzinfo=datetime.timezone.utc)


def test_binary_types(binary_db):
    db = EseDB(binary_db)
    table = db.table("binary")

    assert [(col.name, col.type) for col in table.columns] == [
        ("Id", JET_coltyp.Long),
        ("FixedBinary", JET_coltyp.Binary),
        ("NullableFixedBinary", JET_coltyp.Binary),
        ("Binary", JET_coltyp.Binary),
        ("NullableBinary", JET_coltyp.Binary),
        ("MaxBinary", JET_coltyp.Binary),
        ("TaggedBinary", JET_coltyp.Binary),
        ("NullableTaggedBinary", JET_coltyp.Binary),
        ("LongBinary", JET_coltyp.LongBinary),
        ("LongCompressedBinary", JET_coltyp.LongBinary),
        ("MaxLongBinary", JET_coltyp.LongBinary),
        ("MaxLongCompressedBinary", JET_coltyp.LongBinary),
    ]

    records = list(table.records())
    assert len(records) == 1

    assert records[0].Id == 1
    assert records[0].FixedBinary == b"test fixed binary data" + (b"\x00" * 233)
    assert records[0].NullableFixedBinary is None
    assert records[0].Binary == b"test binary data"
    assert records[0].NullableBinary is None
    assert records[0].MaxBinary == b"test max binary data " + (b"a" * 70)
    assert records[0].TaggedBinary == b"test tagged binary data"
    assert records[0].NullableTaggedBinary is None
    assert records[0].LongBinary == b"test long binary data " + (b"a" * 1000)
    assert records[0].LongCompressedBinary == b"test long compressed binary data " + (b"a" * 1000)
    assert records[0].MaxLongBinary == b"test max long binary data " + (b"a" * 900)
    assert records[0].MaxLongCompressedBinary == b"test max long compressed binary data " + (b"a" * 900)


def test_text_types(text_db):
    db = EseDB(text_db)
    table = db.table("text")

    assert [(col.name, col.type) for col in table.columns] == [
        ("Id", JET_coltyp.Long),
        ("FixedASCII", JET_coltyp.Text),
        ("FixedUnicode", JET_coltyp.Text),
        ("NullableFixedASCII", JET_coltyp.Text),
        ("NullableFixedUnicode", JET_coltyp.Text),
        ("ASCII", JET_coltyp.Text),
        ("Unicode", JET_coltyp.Text),
        ("NullableASCII", JET_coltyp.Text),
        ("NullableUnicode", JET_coltyp.Text),
        ("MaxASCII", JET_coltyp.Text),
        ("MaxUnicode", JET_coltyp.Text),
        ("TaggedASCII", JET_coltyp.Text),
        ("TaggedUnicode", JET_coltyp.Text),
        ("NullableTaggedASCII", JET_coltyp.Text),
        ("NullableTaggedUnicode", JET_coltyp.Text),
        ("LongASCII", JET_coltyp.LongText),
        ("LongUnicode", JET_coltyp.LongText),
        ("LongCompressedASCII", JET_coltyp.LongText),
        ("LongCompressedUnicode", JET_coltyp.LongText),
        ("LongTinyASCII", JET_coltyp.LongText),
        ("LongTinyUnicode", JET_coltyp.LongText),
        ("LongTinyCompressedASCII", JET_coltyp.LongText),
        ("LongTinyCompressedUnicode", JET_coltyp.LongText),
        ("MaxLongASCII", JET_coltyp.LongText),
        ("MaxLongUnicode", JET_coltyp.LongText),
        ("MaxLongCompressedASCII", JET_coltyp.LongText),
        ("MaxLongCompressedUnicode", JET_coltyp.LongText),
    ]

    records = list(table.records())
    assert len(records) == 1

    assert records[0].Id == 1
    assert records[0].FixedASCII == "Fixed ASCII text" + (" " * 239)
    assert records[0].FixedUnicode == "Fixed Unicode text 🦊" + (" " * 107)
    assert records[0].NullableFixedASCII is None
    assert records[0].NullableFixedUnicode is None
    assert records[0].ASCII == "Simple ASCII text"
    assert records[0].Unicode == "Simple Unicode text 🦊"
    assert records[0].NullableASCII is None
    assert records[0].NullableUnicode is None
    assert records[0].MaxASCII == "Max ASCII text that can't be that long"
    assert records[0].MaxUnicode == "Max Unicode text that can't be that long 🦊"
    assert records[0].TaggedASCII == "Tagged ASCII text"
    assert records[0].TaggedUnicode == "Tagged Unicode text 🦊"
    assert records[0].NullableTaggedASCII is None
    assert records[0].NullableTaggedUnicode is None
    assert records[0].LongASCII == "Long ASCII text " + ("a" * 1024)
    assert records[0].LongUnicode == "Long Unicode text 🦊 " + ("a" * 1024)
    assert records[0].LongCompressedASCII == "Long compressed ASCII text " + ("a" * 1024)
    assert records[0].LongCompressedUnicode == "Long compressed Unicode text 🦊 " + ("a" * 1024)
    assert records[0].LongTinyASCII == "Tiny ASCII"
    assert records[0].LongTinyUnicode == "Tiny 🦊"
    assert records[0].LongTinyCompressedASCII == "Tiny c ASCII"
    assert records[0].LongTinyCompressedUnicode == "Tiny c 🦊"
    assert records[0].MaxLongASCII == "Max long ASCII text that can be a bit longer " + ("a" * 900)
    assert records[0].MaxLongUnicode == "Max long Unicode text that can be a bit longer 🦊 " + ("a" * 900)
    assert records[0].MaxLongCompressedASCII == "Max long compressed ASCII text that can be a bit longer " + ("a" * 900)
    assert records[0].MaxLongCompressedUnicode == "Max long compressed Unicode text that can be a bit longer 🦊 " + (
        "a" * 900
    )


def test_multivalue_types(multi_db):
    db = EseDB(multi_db)
    table = db.table("multi")

    assert [(col.name, col.type) for col in table.columns] == [
        ("Id", JET_coltyp.Long),
        ("Bit", JET_coltyp.Bit),
        ("UnsignedByte", JET_coltyp.UnsignedByte),
        ("Short", JET_coltyp.Short),
        ("Long", JET_coltyp.Long),
        ("Currency", JET_coltyp.Currency),
        ("IEEESingle", JET_coltyp.IEEESingle),
        ("IEEEDouble", JET_coltyp.IEEEDouble),
        ("DateTime", JET_coltyp.DateTime),
        ("Binary", JET_coltyp.Binary),
        ("LongBinary", JET_coltyp.LongBinary),
        ("LongCompressedBinary", JET_coltyp.LongBinary),
        ("ASCII", JET_coltyp.Text),
        ("Unicode", JET_coltyp.Text),
        ("LongASCII", JET_coltyp.LongText),
        ("LongUnicode", JET_coltyp.LongText),
        ("LongCompressedASCII", JET_coltyp.LongText),
        ("LongCompressedUnicode", JET_coltyp.LongText),
        ("UnsignedLong", JET_coltyp.UnsignedLong),
        ("LongLong", JET_coltyp.LongLong),
        ("GUID", JET_coltyp.GUID),
        ("UnsignedShort", JET_coltyp.UnsignedShort),
    ]

    records = list(table.records())
    assert len(records) == 2

    assert records[0].Id == 1
    assert records[0].Bit == [False, True]
    assert records[0].UnsignedByte == [0, 127, 255]
    assert records[0].Short == [0, -32767, 32767]
    assert records[0].Long == [0, -2147483647, 2147483647]
    assert records[0].Currency == [0, -9223372036854775807, 9223372036854775807]
    assert records[0].IEEESingle == [0.0, -1.0, 1.0]
    assert records[0].IEEEDouble == [0.0, -1.0, 1.0]
    assert list(map(oatimestamp, records[0].DateTime)) == [
        datetime.datetime(1661, 4, 17, 11, 30, tzinfo=datetime.timezone.utc),
        datetime.datetime(2077, 4, 1, 0, 0, tzinfo=datetime.timezone.utc),
        datetime.datetime(2517, 9, 24, 5, 30, tzinfo=datetime.timezone.utc),
    ]
    assert records[0].Binary == [
        b"Some binary data that has multiple values, this is value 1",
        b"Some binary data that has multiple values, this is value 2",
        b"Some binary data that has multiple values, this is value 3",
    ]
    assert records[0].LongBinary == [
        b"Some very long binary data that has multiple values, this is value 1 " + (b"a" * 1024),
        b"Some very long binary data that has multiple values, this is value 2 " + (b"a" * 1024),
        b"Some very long binary data that has multiple values, this is value 3 " + (b"a" * 1024),
    ]
    assert records[0].LongCompressedBinary == [
        b"Some very long compressed binary data that has multiple values, this is value 1 " + (b"a" * 1024),
        b"Some very long compressed binary data that has multiple values, this is value 2 " + (b"a" * 1024),
        b"Some very long compressed binary data that has multiple values, this is value 3 " + (b"a" * 1024),
    ]
    assert records[0].ASCII == [
        "Some ASCII text that has multiple values, this is value 1",
        "Some ASCII text that has multiple values, this is value 2",
        "Some ASCII text that has multiple values, this is value 3",
    ]
    assert records[0].Unicode == [
        "Some Unicode text that has multiple values, this is value 1 🦊",
        "Some Unicode text that has multiple values, this is value 2 🦊🦊",
        "Some Unicode text that has multiple values, this is value 3 🦊🦊🦊",
    ]
    assert records[0].LongASCII == [
        "Some very long ASCII text that has multiple values, this is value 1 " + ("a" * 1024),
        "Some very long ASCII text that has multiple values, this is value 2 " + ("a" * 1024),
        "Some very long ASCII text that has multiple values, this is value 3 " + ("a" * 1024),
    ]
    assert records[0].LongUnicode == [
        "Some very long Unicode text that has multiple values, this is value 1 🦊 " + ("a" * 1024),
        "Some very long Unicode text that has multiple values, this is value 2 🦊🦊 " + ("a" * 1024),
        "Some very long Unicode text that has multiple values, this is value 3 🦊🦊🦊 " + ("a" * 1024),
    ]
    assert records[0].LongCompressedASCII == [
        "Some very long compressed ASCII text that has multiple values, this is value 1 " + ("a" * 1024),
        "Some very long compressed ASCII text that has multiple values, this is value 2 " + ("a" * 1024),
        "Some very long compressed ASCII text that has multiple values, this is value 3 " + ("a" * 1024),
    ]
    assert records[0].LongCompressedUnicode == [
        "Some very long compressed Unicode text that has multiple values, this is value 1 🦊 " + ("a" * 1024),
        "Some very long compressed Unicode text that has multiple values, this is value 2 🦊🦊 " + ("a" * 1024),
        "Some very long compressed Unicode text that has multiple values, this is value 3 🦊🦊🦊 " + ("a" * 1024),
    ]
    assert records[0].UnsignedLong == [0, 4294967295]
    assert records[0].LongLong == [0, -9223372036854775807, 9223372036854775807]
    assert records[0].GUID == [
        "03402861-fad3-4ce5-986e-d31df852f2a7",
        "09b589e3-a92b-4936-bbc4-bb9dff334bf3",
        "2212ff6a-6712-4fe4-bb4a-21e6d0e043d1",
    ]
    assert records[0].UnsignedShort == [0, 65535]

    # Test some small "long" multi-values and two-values
    # Also test that the other columns should correctly be set to None
    assert records[1].Id == 2
    assert records[1].Bit is None
    assert records[1].UnsignedByte is None
    assert records[1].Short is None
    assert records[1].Long is None
    assert records[1].Currency is None
    assert records[1].IEEESingle is None
    assert records[1].IEEEDouble is None
    assert records[1].DateTime is None
    assert records[1].Binary is None
    assert records[1].LongBinary == [
        b"Tiny binary 1",
        b"Tiny binary 2",
        b"Tiny binary 3",
    ]
    assert records[1].LongCompressedBinary == [
        b"Tiny c binary 1",
        b"Tiny c binary 2",
        b"Tiny c binary 3",
    ]
    assert records[1].ASCII is None
    assert records[1].Unicode is None
    assert records[1].LongASCII == [
        "Tiny ASCII 1",
        "Tiny ASCII 2",
    ]
    assert records[1].LongUnicode == [
        "Tiny 🦊 1",
        "Tiny 🦊🦊",
        "Tiny 🦊🦊🦊",
    ]
    assert records[1].LongCompressedASCII == [
        "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
        "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        "ccccccccccccccccccccccccccccccccccc",
    ]
    assert records[1].LongCompressedUnicode == [
        "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa 🦊",
        "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb 🦊🦊",
    ]
    assert records[1].UnsignedLong is None
    assert records[1].LongLong is None
    assert records[1].GUID is None
    assert records[1].UnsignedShort is None


def test_default_db(default_db):
    db = EseDB(default_db)
    table = db.table("default")

    records = list(table.records())
    assert len(records) == 1

    assert records[0].Id == 1
    assert records[0].Bit is True
    assert records[0].UnsignedByte == 69
    assert records[0].Short == 0x1234
    assert records[0].Long == 0x12345678
    assert records[0].Currency == 0x123456789ABCDEF0
    assert records[0].IEEESingle == 1.0
    assert records[0].IEEEDouble == 2.0
    assert oatimestamp(records[0].DateTime) == datetime.datetime(2022, 10, 4, 0, 0, tzinfo=datetime.timezone.utc)
    assert records[0].UnsignedLong == 12345678
    assert records[0].LongLong == 0xC001DEADD00D
    assert records[0].GUID == "c001d00d-dead-beef-face-feeddeadbeef"
    assert records[0].UnsignedShort == 0xF00D

    assert records[0].Binary == b"Short default binary"
    assert records[0].LongBinary == b"Long default binary " + (b"a" * 200)
    assert records[0].ASCII == "Short default ASCII"
    assert records[0].Unicode == "Short default Unicode 🦊"
    assert records[0].LongASCII == "Long default ASCII " + ("a" * 200)
    assert records[0].LongUnicode == "Long default Unicode 🦊 " + ("a" * 64)


def test_large_db(large_db):
    db = EseDB(large_db)
    table = db.table("large")

    assert db.page_size == 32 * 1024
    assert [(col.name, col.type) for col in table.columns] == [
        ("Id", JET_coltyp.Long),
    ] + [(f"Column{i}", JET_coltyp.UnsignedShort) for i in range(64993)]

    records = list(table.records())
    assert len(records) == 16

    for i in range(64993):
        assert records[i // 4096].get(f"Column{i}") == i



# Extensible Storage Engine (ESE) Database implementation
# Combination of pre-source release reverse engineering and post-source release cleanup
# Reference: https://github.com/microsoft/Extensible-Storage-Engine

from functools import cached_property, lru_cache
from typing import BinaryIO, Iterator

from dissect.esedb.c_esedb import c_esedb, pgnoFDPMSO, ulDAEMagic
from dissect.esedb.exceptions import InvalidDatabase
from dissect.esedb.page import Page
from dissect.esedb.table import Catalog, Table


class EseDB:
    """EseDB class.

    Loads an ESE database from the given file handle. Optionally enable impacket compatible data output.

    Impacket compatibility limits what values are parsed and returns most values as a hex string. Most notably,
    long and multi values are not parsed.

    Args:
        fh: The file-like object to open an ESE database on.
        impacket_compat: Whether to make the output impacket compatible.

    Raises:
        InvalidDatabase: If the file-like object does not look like an ESE database.
    """

    def __init__(self, fh: BinaryIO, impacket_compat: bool = False):
        self.fh = fh
        self.impacket_compat = impacket_compat

        self.header = c_esedb.DBFILEHDR(fh)
        if self.header.ulMagic != ulDAEMagic:
            raise InvalidDatabase("invalid file header signature")

        self.page_size = self.header.cbPageSize
        self.version = self.header.ulVersion
        self.format_major = self.header.ulDaeUpdateMajor
        self.format_minor = self.header.ulDaeUpdateMinor

        if self.format_major < 9:
            raise InvalidDatabase("unsupported format revision")

        self.catalog = Catalog(self, pgnoFDPMSO)

    @cached_property
    def has_small_pages(self) -> bool:
        """Return whether this database has small pages (<= 8K)."""
        return self.page_size <= 8192

    def table(self, name: str) -> Table:
        """Get a table by name.

        Args:
            name: The table to retrieve.
        """
        return self.catalog.table(name)

    def tables(self) -> list[Table]:
        """Get a list of all tables."""
        return self.catalog.tables

    def read_page(self, num: int) -> bytes:
        """Get the physical page data.

        Args:
            num: The physical page number to retrieve.

        Raises:
            IndexError: If the page number is out of bounds.
        """
        if num < 1:
            raise IndexError("page number cannot be less than 1")

        self.fh.seek((num - 1) * self.page_size)
        buf = self.fh.read(self.page_size)

        if len(buf) != self.page_size:
            raise IndexError("page number exceeds file size")

        return buf

    @lru_cache(maxsize=4096)
    def page(self, num: int) -> Page:
        """Get a logical page.

        The first two pages in the file are the Header and Shadow Header pages.
        Logical pages start at physical page 2 (zero-indexed).

        Args:
            num: The logical page number to retrieve.
        """
        buf = self.read_page(num + 2)
        return Page(self, num, buf)

    def pages(self) -> Iterator[Page]:
        """Iterate over all pages."""
        num = 1
        while True:
            try:
                yield self.page(num)
                num += 1
            except IndexError:
                break
